@@ -50,7 +50,7 @@ NOTE: The statements sent to the database system can be configured, but this
 NOTE: should be made convenient in future releases.
 """
 
-__version__ = "0.2"
+__version__ = "0.2.1"
 
 import sys
 import time, random # for sqlite support
@@ -74,7 +74,7 @@ class URIRef(Special):
 class Literal(Special):
     pass
 
-class Namespace(Special):
+class Namespace(URIRef):
     def __getitem__(self, item):
         try:
             return unicode(self)[item]
@@ -94,6 +94,58 @@ class BNode:
         # NOTE: Should probably be more careful to avoid collisions with non-BNodes.
         return hash(self.value)
 
+# Special pattern classes.
+
+class AbstractExpression:
+    pass
+
+class Expression(AbstractExpression):
+
+    # NOTE: Improve the object type mechanism.
+
+    def __init__(self, expr, values, object_type="L"):
+        self.expr = expr
+        self.values = values
+        self.object_type = object_type
+
+    def __str__(self):
+        return self.expr.replace("_", "%s")
+
+    def __len__(self):
+        return 2
+
+    def __getitem__(self, i):
+        return (self.expr, self.values)[i]
+
+class Conjunction(AbstractExpression):
+    def __init__(self, values):
+        self.values = values
+
+    def __str__(self):
+        return " and ".join(["%s"] * len(self.values))
+
+class Disjunction(AbstractExpression):
+    def __init__(self, values):
+        self.values = values
+
+    def __str__(self):
+        return " or ".join(["%s"] * len(self.values))
+
+class Pattern:
+    def __init__(self, subject, predicate, object):
+        self.subject = subject
+        self.predicate = predicate
+        self.object = object
+
+    def __len__(self):
+        return 3
+
+    def __getitem__(self, i):
+        return (self.subject, self.predicate, self.object)[i]
+
+class Defined:
+    pass
+
 # Useful defaults.
 
 RDFNS = Namespace("http://www.w3.org/1999/02/22-rdf-syntax-ns#")
@@ -106,7 +158,7 @@ class Adapter:
     # A table indicating whether a paramstyle mandates a dictionary when values
     # are being presented to the database system.
 
-    paramdict = {
+    _paramdict = {
         "qmark" : 0, "numeric" : 0, "named" : 1, "format" : 0, "pyformat" : 1
         }
 
@@ -128,7 +180,7 @@ class Adapter:
             return "?" # NOTE: Guess a paramstyle.
 
     def _present(self, values):
-        if self.paramdict.get(self.paramstyle, 0):
+        if self._paramdict.get(self.paramstyle, 0):
             d = {}
             i = 0
             for value in values:
@@ -244,6 +296,18 @@ class TripleStore:
         self.sequence_name = sequence_name or "bnode"
         self.debug = debug
 
+        # Useful classes.
+
+        self.Expression = Expression
+        self.Conjunction = Conjunction
+        self.Disjunction = Disjunction
+        self.Pattern = Pattern
+        self.Defined = Defined
+
+        # Special flags.
+
+        self.supports_querying = 1
+
     def get_context(self, context):
 
         """
@@ -375,7 +439,9 @@ class TripleStore:
 
         # Subjects are always tested for equality.
 
-        if subject is not None:
+        if isinstance(subject, self.Defined):
+            conditions.append("%s is not null" % self._column("subject", index))
+        elif subject is not None:
             subject_value, subject_type = self._convert(subject)
             conditions.append("%s = ?" % self._column("subject", index))
             values.append(subject_value)
@@ -384,9 +450,19 @@ class TripleStore:
 
         # Predicates are always tested for equality.
 
-        if predicate is not None:
-            conditions.append("%s = ?" % self._column("predicate", index))
-            values.append(unicode(predicate))
+        if isinstance(predicate, self.Defined):
+            conditions.append("%s is not null" % self._column("predicate", index))
+        elif predicate is not None:
+            if isinstance(predicate, list):
+                conditions.append("%s in (%s)" % (
+                    self._column("predicate", index),
+                    ", ".join(["?"] * len(predicate))
+                    ))
+                for p in predicate:
+                    values.append(unicode(p))
+            else:
+                conditions.append("%s = ?" % self._column("predicate", index))
+                values.append(unicode(predicate))
 
         # Objects can either be single values tested for equality, Results
         # objects describing some set of already selected results, or more
@@ -394,20 +470,28 @@ class TripleStore:
 
         new_index = index # the current table index - may be changed by nested queries
 
-        if object is not None:
+        if isinstance(object, self.Defined):
+            conditions.append("%s is not null" % self._column("object", index))
+        elif object is not None:
 
             # Process an expression and values.
 
-            if isinstance(object, tuple):
+            if isinstance(object, Expression) or isinstance(object, tuple):
                 comparison, comparison_values = object
                 conditions.append(comparison.replace("_", self._column("object", index)))
                 values += map(unicode, comparison_values)
                 conditions.append("%s = ?" % self._column("object_type", index))
-                values.append("L")
 
-            # Process a Results object.
+                # NOTE: Improve this mechanism.
 
-            elif isinstance(object, Results):
+                if hasattr(object, "object_type"):
+                    values.append(object.object_type)
+                else:
+                    values.append("L")
+
+            # Process a Query object.
+
+            elif isinstance(object, Query):
                 subquery, subquery_values, new_index = object.get_child_query(index)
                 conditions.append(subquery)
                 values += subquery_values
@@ -515,27 +599,10 @@ class TripleStore:
 
     # Querying and updating methods compatible with rdflib.
 
-    def subjects(self, predicate=None, object=None, ordering=None, limit=None):
-
-        "Return subjects for the given 'predicate' and/or 'object' identifiers."
-
-        return Subjects(self, (None, predicate, object), ordering=ordering, limit=limit)
-
-    def predicates(self, subject=None, object=None, ordering=None, limit=None):
-
-        "Return predicates for the given 'subject' and/or 'object' identifiers."
-
-        return Predicates(self, (subject, None, object), ordering=ordering, limit=limit)
-
-    def objects(self, subject=None, predicate=None, fn=None, ordering=None, limit=None):
+    def subjects(self, predicate=None, object=None, pattern=None, ordering=None, limit=None):
 
         """
-        Return objects for the given 'subject' and/or 'predicate' identifiers.
-        Where the given function 'fn' is specified (as a string containing a
-        reference to selected objects as denoted by the "_" character), this
-        function is applied the underlying query; for example:
-
-        "substring(_ from 1 for 4)"
+        Return subjects for the given 'predicate' and/or 'object' identifiers.
 
         Where the 'ordering' is specified (as a tuple optionally containing
         predicates and ending with a ordering "direction" - either "asc" or
@@ -549,13 +616,85 @@ class TripleStore:
         be limited to the stated amount.
         """
 
-        return Objects(self, (subject, predicate, None), fn=fn, ordering=ordering, limit=limit)
+        if pattern is not None:
+            return Subjects(self, pattern, ordering=ordering, limit=limit)
+        else:
+            return Subjects(self, (None, predicate, object), ordering=ordering, limit=limit)
 
-    def triples(self, pattern, ordering=None, limit=None):
+    def predicates(self, subject=None, object=None, pattern=None, ordering=None, limit=None):
 
-        "Return triples conforming to the given 'pattern'."
+        """
+        Return predicates for the given 'subject' and/or 'object' identifiers.
 
-        return Triples(self, pattern, ordering=ordering, limit=limit)
+        Where the 'ordering' is specified (as a tuple optionally containing
+        predicates and ending with a ordering "direction" - either "asc" or
+        "desc"), a query is produced which attempts to find objects through the
+        traversal of triples via the predicates and to sort them accordingly;
+        for example:
+
+        ("pr1", "pr2", "asc")
+
+        Where a 'limit' is specified (as an integer), the number of results will
+        be limited to the stated amount.
+        """
+
+        if pattern is not None:
+            return Predicates(self, pattern, ordering=ordering, limit=limit)
+        else:
+            return Predicates(self, (subject, None, object), ordering=ordering, limit=limit)
+
+    def objects(self, subject=None, predicate=None, pattern=None, fn=None, ordering=None, limit=None):
+
+        """
+        Return objects for the given 'subject' and/or 'predicate' identifiers.
+
+        Where the given function 'fn' is specified (as a string containing a
+        reference to selected objects as denoted by the "_" character), this
+        function is applied in the underlying query; for example:
+
+        "substr(_, 1, 4)"
+
+        Where the 'ordering' is specified (as a tuple optionally containing
+        predicates and ending with a ordering "direction" - either "asc" or
+        "desc"), a query is produced which attempts to find objects through the
+        traversal of triples via the predicates and to sort them accordingly;
+        for example:
+
+        ("pr1", "pr2", "asc")
+
+        Where a 'limit' is specified (as an integer), the number of results will
+        be limited to the stated amount.
+        """
+
+        if pattern is not None:
+            return Objects(self, pattern, fn=fn, ordering=ordering, limit=limit)
+        else:
+            return Objects(self, (subject, predicate, None), fn=fn, ordering=ordering, limit=limit)
+
+    def triples(self, pattern, fn=None, ordering=None, limit=None):
+
+        """
+        Return triples conforming to the given 'pattern'.
+
+        Where the given function 'fn' is specified (as a string containing a
+        reference to selected objects as denoted by the "_" character), this
+        function is applied in the underlying query; for example:
+
+        "substr(_, 1, 4)"
+
+        Where the 'ordering' is specified (as a tuple optionally containing
+        predicates and ending with a ordering "direction" - either "asc" or
+        "desc"), a query is produced which attempts to find objects through the
+        traversal of triples via the predicates and to sort them accordingly;
+        for example:
+
+        ("pr1", "pr2", "asc")
+
+        Where a 'limit' is specified (as an integer), the number of results will
+        be limited to the stated amount.
+        """
+
+        return Triples(self, pattern, fn=fn, ordering=ordering, limit=limit)
 
     def add(self, pattern):
 
@@ -613,6 +752,17 @@ class TripleStore:
         s = self.get_context(context)
         s.remove((None, None, None))
 
+    # Additional methods.
+
+    def intersection(self, queries):
+        return Intersection(self, queries)
+
+    def union(self, queries):
+        return Union(self, queries)
+
+    def negation(self, result):
+        return Negation(self, result)
+
 # Convenience classes.
 
 class Subject:
@@ -695,7 +845,54 @@ class Subject:
 
 # Querying classes.
 
-class Results:
+class Query:
+
+    "An abstract class providing support for lazy query evaluation."
+
+    def __init__(self, store):
+        self.store = store
+        self.results = None
+        self.result_type = self
+
+    def _ensure(self):
+        if self.results is None:
+            self.results = self._execute()
+
+    def __getitem__(self, i):
+        self._ensure()
+        return self.results[i]
+
+    def __len__(self):
+        self._ensure()
+        return len(self.results)
+
+    def __repr__(self):
+        self._ensure()
+        return repr(self.results)
+
+    def get_child_query(self, index):
+
+        """
+        Obtain the query clause and values (as a 2-tuple) which links this
+        object's query to a parent query. Employ the given 'index' to correctly
+        qualify the clause's "exposed" columns.
+        """
+
+        query = "select * from %s as %s" % (
+            self.store._table(), self.store._table(index + 1)
+            )
+        constraints = "%s = %s and %s = %s and %s = %s" % (
+            self.store._column("object", index),
+            self.store._column("subject", index + 1),
+            self.store._column("object_type", index),
+            self.store._column("subject_type", index + 1),
+            self.store._column("context", index),
+            self.store._column("context", index + 1)
+            )
+        query, values, new_index = self._get_query(query, index + 1, constraints)
+        return "exists (%s)" % query, values, new_index
+
+class Results(Query):
 
     """
     A class representing a selection of results.
@@ -722,30 +919,13 @@ class Results:
         optional 'ordering' description and optional 'limit' criteria.
         """
 
-        self.store = store
         self.pattern = pattern
         self.fn = fn
         ordering = ordering or [None]
         self.order_fields = ordering[:-1]
         self.order_direction = ordering[-1]
         self.limit = limit
-        self.results = None
-
-    def _ensure(self):
-        if self.results is None:
-            self.results = self._execute()
-
-    def __getitem__(self, i):
-        self._ensure()
-        return self.results[i]
-
-    def __len__(self):
-        self._ensure()
-        return len(self.results)
-
-    def __repr__(self):
-        self._ensure()
-        return repr(self.results)
+        Query.__init__(self, store)
 
     def _get_query(self, select_clause, index, constraints=None):
 
@@ -773,8 +953,26 @@ class Results:
 
         # Add general conditions.
 
-        if isinstance(self.pattern, tuple):
+        if isinstance(self.pattern, Pattern) or isinstance(self.pattern, tuple):
             conditions, values, new_index = self.store._get_conditions(self.pattern, index=index)
+
+        # For expressions rather than triples, prepare a query which includes
+        # all the subqueries.
+
+        elif isinstance(self.pattern, AbstractExpression):
+            conditions = []
+            values = []
+            subconditions = []
+            new_index = index
+            for t in self.pattern.values:
+                subquery, subquery_values, new_index = t.get_subquery(index, new_index)
+                subconditions.append(subquery)
+                values += subquery_values
+            if len(subconditions) != 0:
+                conditions.append(str(self.pattern) % tuple(subconditions))
+
+        # For other things, add each subquery as a condition.
+
         else:
             conditions = []
             values = []
@@ -837,28 +1035,6 @@ class Results:
             this_index = "order_%s" % (index + i)
         return self.store._column("object", this_index)
 
-    def get_child_query(self, index):
-
-        """
-        Obtain the query clause and values (as a 2-tuple) which links this
-        object's query to a parent query. Employ the given 'index' to correctly
-        qualify the clause's "exposed" columns.
-        """
-
-        query = "select * from %s as %s" % (
-            self.store._table(), self.store._table(index + 1)
-            )
-        constraints = "%s = %s and %s = %s and %s = %s" % (
-            self.store._column("object", index),
-            self.store._column("subject", index + 1),
-            self.store._column("object_type", index),
-            self.store._column("subject_type", index + 1),
-            self.store._column("context", index),
-            self.store._column("context", index + 1)
-            )
-        query, values, new_index = self._get_query(query, index + 1, constraints)
-        return "exists (%s)" % query, values, new_index
-
     def _apply_function(self, fn, column):
         if fn is not None:
             return fn.replace("_", column)
@@ -1046,8 +1222,8 @@ class Objects(SingleResults):
         """
 
         constraints = "%s = %s and %s = %s and %s = %s" % (
-            self.store._column("object", parent_index),
-            self.store._column("object", index + 1),
+            self._apply_function(self.fn, self.store._column("object", parent_index)),
+            self._apply_function(self.fn, self.store._column("object", index + 1)),
             self.store._column("object_type", parent_index),
             self.store._column("object_type", index + 1),
             self.store._column("context", parent_index),
@@ -1087,7 +1263,7 @@ class Triples(Results):
         query = "select %s, %s, %s, %s, %s, %s from %s as %s" % (
             self.store._column("subject", index),
             self.store._column("predicate", index),
-            self.store._column("object", index),
+            self._apply_function(self.fn, self.store._column("object", index)),
             self.store._column("subject_type", index),
             self.store._column("object_type", index),
             self.store._column("context", index),
@@ -1109,8 +1285,8 @@ class Triples(Results):
             self.store._column("subject", index + 1),
             self.store._column("predicate", parent_index),
             self.store._column("predicate", index + 1),
-            self.store._column("object", parent_index),
-            self.store._column("object", index + 1),
+            self._apply_function(self.fn, self.store._column("object", parent_index)),
+            self._apply_function(self.fn, self.store._column("object", index + 1)),
             self.store._column("subject_type", parent_index),
             self.store._column("subject_type", index + 1),
             self.store._column("object_type", parent_index),
@@ -1122,6 +1298,90 @@ class Triples(Results):
         query = "exists (%s)" % internal_query
         return query, values, new_index
 
+class Negation(Query):
+    def __init__(self, store, result):
+        self.result = result
+        Query.__init__(self, store)
+
+    def get_subquery(self, parent_index, index):
+        query, values, new_index = self.result.get_subquery(parent_index, index)
+        return "not " + query, values, new_index
+
+    # NOTE: Consider implementing this!
+    #def get_query(self, index):
+    #    return self.result.get_query(index)
+
+    def _execute(self):
+        return self.result._execute()
+
+    def get_child_query(self, index):
+        query, values, new_index = self.query.get_child_query(index)
+        return "not " + query, values, new_index
+
+class SetQuery(Query):
+
+    "A class representing a set operation on query results."
+
+    def __init__(self, store, operation, queries):
+        self.operation = " %s " % operation
+        self.queries = queries
+        Query.__init__(self, store)
+        if len(self.queries) > 0:
+            self.result_type = self.queries[0].result_type
+        else:
+            self.result_type = None
+
+    def _execute(self):
+        cursor = self.store.connection.cursor()
+        try:
+            final_query, values, new_index = self.get_query(0)
+            if self.store.debug:
+                print self.store._pmarks(final_query), self.store._present(values)
+            if values:
+                cursor.execute(self.store._pmarks(final_query), self.store._present(values))
+            else:
+                cursor.execute(self.store._pmarks(final_query))
+            results = cursor.fetchall()
+        finally:
+            cursor.close()
+
+        if isinstance(self.result_type, Triples):
+            return [(self.store._instantiate(s, st), URIRef(p), self.store._instantiate(o, ot)) for (s, p, o, st, ot, c) in results]
+        elif isinstance(self.result_type, SingleResults):
+            return [self.store._instantiate(row[0], row[1]) for row in results]
+        else:
+            return []
+
+    def get_query(self, index):
+        parts = []
+        values = []
+        for query in self.queries:
+            query_text, query_values, new_index = query.get_query(index)
+            parts.append(query_text)
+            values += query_values
+
+        # NOTE: sqlite doesn't seem to support more than simple left-to-right
+        # NOTE: evaluation of union/intersect with select statements. Thus, we
+        # NOTE: have to omit the parentheses below, meaning that hierarchies of
+        # NOTE: SetQuery objects may not yield the expected results.
+
+        final_query = self.operation.join(parts) # "(%s)" % self.operation.join(parts)
+        return final_query, values, new_index
+
+class Intersection(SetQuery):
+
+    "A class representing the intersection of a number of queries."
+
+    def __init__(self, store, queries):
+        SetQuery.__init__(self, store, "intersect", queries)
+
+class Union(SetQuery):
+
+    "A class representing the union of a number of queries."
+
+    def __init__(self, store, queries):
+        SetQuery.__init__(self, store, "union", queries)
+
 # Convenience functions.
 
 def open(database_name=None, database_module_name=None, adapter=None, table_name=None, sequence_name=None, debug=0, **kw):
